"""
Model data schemas and classes.
"""

from dataclasses import dataclass
from typing import Type

from marshmallow import fields, post_load

from .json_schema import *


class ModelDataSchema(JsonSchema):
    name = fields.String(data_key="name", required=False)
    version = fields.String(data_key="version", required=False)
    authors = fields.List(fields.String(), data_key="authors", required=False)
    description = fields.String(data_key="description", required=False)
    comment = fields.String(data_key="comment", required=False)
    doi = fields.String(data_key="doi", required=False)
    url = fields.String(data_key="url", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> "ModelData":
        obj = super().make_object(data, **kwargs)
        return ModelData(
            name=obj.name,
            version=obj.version,
            authors=obj.authors,
            description=obj.description,
            comment=obj.comment,
            doi=obj.doi,
            url=obj.url,
        )


@dataclass
class ModelData(JsonSchemaResult):
    name: str | None = None
    version: str | None = None
    authors: list[str] | None = None
    description: str | None = None
    comment: str | None = None
    doi: str | None = None
    url: str | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return ModelDataSchema



"""
Variable valuation schemas and classes.
"""

from dataclasses import dataclass

from marshmallow import fields, post_load, validate
from marshmallow_oneofschema.one_of_schema import OneOfSchema

import umbi.datatypes

from .json_schema import *


class ValuationPaddingSchema(JsonSchema):
    """Schema for padding fields."""

    padding = FieldUint(data_key="padding", required=True)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructPadding:
        """Create a Padding object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return umbi.datatypes.StructPadding(padding=obj.padding)


class ValuationAttributeSchema(JsonSchema):
    """Schema for variable fields."""

    name = fields.String(data_key="name", required=True)
    type = fields.String(
        data_key="type", required=True, validate=validate.OneOf(["bool", "int", "uint", "double", "rational", "string"])
    )
    size = FieldUint(data_key="size", required=False)
    lower = fields.Float(data_key="lower", required=False)
    upper = fields.Float(data_key="upper", required=False)
    offset = fields.Float(data_key="offset", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructAttribute:
        """Validate and create a Variable object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        # Check for unsupported fields
        for field in ("lower", "upper", "offset"):
            if getattr(obj, field, None) is not None:
                raise NotImplementedError(
                    f"feature not implemented: '{field}' must be omitted, but got {getattr(obj, field)}."
                )

        return umbi.datatypes.StructAttribute(
            name=obj.name,
            type=obj.type,
            size=getattr(obj, "size", None),
            lower=getattr(obj, "lower", None),
            upper=getattr(obj, "upper", None),
            offset=getattr(obj, "offset", None),
        )


class ValuationFieldSchema(OneOfSchema):
    """Schema for valuation fields (padding or variable)."""

    type_schemas = {
        "padding": ValuationPaddingSchema,
        "attribute": ValuationAttributeSchema,
    }

    # custom discriminator field since the default 'type' field is used in ValuationAttributeSchema
    type_field = "_discriminator"

    def get_obj_type(self, obj):
        """Determine which schema to use based on the object's attributes."""
        if hasattr(obj, "padding"):
            return "padding"
        elif hasattr(obj, "name") and hasattr(obj, "type"):
            return "attribute"
        else:
            raise ValueError("Object must be either a padding or attribute namespace")

    def load(self, json_data, *args, **kwargs):
        """Add discriminator field before loading."""
        assert isinstance(json_data, dict)
        json_data = dict(json_data, _discriminator="padding" if "padding" in json_data else "attribute")
        return super().load(json_data, *args, **kwargs)

    def dump(self, obj, *args, **kwargs):
        """Remove discriminator field after dumping."""
        result = super().dump(obj, *args, **kwargs)
        assert isinstance(result, dict)
        result.pop("_discriminator", None)
        return result


class VariableValuationsSchema(JsonSchema):
    """Schema for variable valuations."""

    alignment = FieldUint(data_key="alignment", required=True)
    variables = fields.List(fields.Nested(ValuationFieldSchema), data_key="variables", required=True)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructType:
        obj = super().make_object(data, **kwargs)
        return umbi.datatypes.StructType(
            alignment=obj.alignment,
            fields=obj.variables,
        )

    def dump(self, obj, *args, **kwargs):
        assert isinstance(obj, umbi.datatypes.StructType)
        obj_dict = {
            "alignment": obj.alignment,
            "variables": self.fields["variables"].serialize("variables", {"variables": obj.fields}),
        }
        return obj_dict



"""Utilities for (de)serializing intervals."""

from fractions import Fraction

import umbi.datatypes
from umbi.datatypes import CommonType, Interval, interval_base_type

from .floats import bytes_to_double, double_to_bytes
from .rationals import bytes_to_rational, num_bytes_for_rational, rational_to_bytes


def interval_to_bytes(interval: Interval, type: CommonType, little_endian: bool = True) -> bytes:
    """Convert an Interval into a bytestring representation."""
    umbi.datatypes.assert_interval_type(type)
    base_type = interval_base_type(type)

    # Updated references to use directly imported symbols.
    converter = None
    if base_type == CommonType.RATIONAL:
        assert isinstance(interval.left, Fraction) and isinstance(interval.right, Fraction)
        # ensure both rationals use the same size for numerator and denominator
        term_size = max(num_bytes_for_rational(interval.left), num_bytes_for_rational(interval.right)) // 2
        converter = lambda value: rational_to_bytes(value, term_size, little_endian)
    else:
        assert base_type == CommonType.DOUBLE
        assert isinstance(interval.left, float) and isinstance(interval.right, float)
        converter = lambda value: double_to_bytes(value, little_endian)
    lower = converter(interval.left)
    upper = converter(interval.right)
    return lower + upper


def bytes_to_interval(data: bytes, type: CommonType, little_endian: bool = True) -> Interval:
    """Convert a bytestring representing an interval into an Interval object."""
    assert len(data) % 2 == 0, "interval data must have even length"
    mid = len(data) // 2
    base_type = interval_base_type(type)
    lower, upper = data[:mid], data[mid:]
    converter = {
        CommonType.DOUBLE: bytes_to_double,
        CommonType.RATIONAL: bytes_to_rational,
    }[base_type]
    lower = converter(lower, little_endian)
    upper = converter(upper, little_endian)
    return Interval(lower, upper)



"""
umbi.ats: Collection of interfaces over a umbfile.
"""

from .annotation import Annotation, AtomicPropositionAnnotation, RewardAnnotation
from .explicit_ats import ExplicitAts, TimeType
from .model_info import ModelInfo

__all__ = [
    "ModelInfo",
    "Annotation",
    "RewardAnnotation",
    "AtomicPropositionAnnotation",
    "TimeType",
    "ExplicitAts",
]



"""
Annotation schemas and classes.
"""

from dataclasses import dataclass
from typing import Literal, Type

from marshmallow import fields, post_load, validate

from .json_schema import *


class AnnotationSchema(JsonSchema):
    """Schema for one annotation."""

    alias = fields.String(data_key="alias", required=False)
    description = fields.String(data_key="description", required=False)
    applies_to = fields.List(
        fields.String(validate=validate.OneOf(["states", "choices", "branches"])),
        data_key="applies-to",
        required=False,
        validate=validate.Length(min=1),
    )
    type = fields.String(
        data_key="type",
        required=False,
        validate=validate.OneOf(["bool", "double", "rational", "double-interval", "rational-interval", "string"]),
    )
    lower = fields.Float(data_key="lower", required=False)
    upper = fields.Float(data_key="upper", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> "Annotation":
        """Create an Annotation object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return Annotation(
            alias=obj.alias,
            description=obj.description,
            applies_to=obj.applies_to,
            type=obj.type,
            lower=obj.lower,
            upper=obj.upper,
        )


@dataclass
class Annotation(JsonSchemaResult):
    """Annotation data class."""

    alias: str | None = None
    description: str | None = None
    applies_to: list[Literal["states", "choices", "branches"]] | None = None
    type: Literal["bool", "double", "rational", "double-interval", "rational-interval", "string"] | None = None
    lower: float | None = None
    upper: float | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return AnnotationSchema


class AnnotationsSchema(JsonSchema):
    """Schema for all annotations."""

    rewards = fields.Dict(
        keys=fields.String(), values=fields.Nested(AnnotationSchema), data_key="rewards", required=False
    )
    aps = fields.Dict(keys=fields.String(), values=fields.Nested(AnnotationSchema), data_key="aps", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> "Annotations":
        """Create an Annotations object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return Annotations(
            rewards=obj.rewards,
            aps=obj.aps,
        )


@dataclass
class Annotations(JsonSchemaResult):
    """Annotations data class."""

    rewards: dict[str, Annotation] | None = None
    aps: dict[str, Annotation] | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return AnnotationsSchema



from fractions import Fraction

from .common_type import (
    CommonType,
    interval_base_type,
    is_integer_type,
    is_interval_type,
)
from .interval import Interval
from .json import is_json_instance
from .struct import StructType

""" Alias for primitive numeric types. """
NumericPrimitive = int | float | Fraction
""" Alias for all numeric types, including intervals. """
Numeric = NumericPrimitive | Interval


def get_instance_type(value: object) -> CommonType:
    """Determine the common type of a given value."""
    if isinstance(value, bool):
        return CommonType.BOOLEAN
    elif isinstance(value, int):
        return CommonType.INT
    elif isinstance(value, float):
        return CommonType.DOUBLE
    elif isinstance(value, Fraction):
        return CommonType.RATIONAL
    elif isinstance(value, Interval):
        if isinstance(value.left, Fraction) or isinstance(value.right, Fraction):
            return CommonType.RATIONAL_INTERVAL
        else:
            return CommonType.DOUBLE_INTERVAL
    elif isinstance(value, str):
        return CommonType.STRING
    elif is_json_instance(value):
        return CommonType.JSON
    elif isinstance(value, StructType):
        return CommonType.STRUCT
    else:
        raise ValueError(f"cannot match value to a common type: {value}")


def is_instance_of_common_type(value: object, type: CommonType) -> bool:
    """Check if a value is an instance of the given common type."""
    value_type = get_instance_type(value)
    if value_type == CommonType.INT:
        # int matches all integer types
        # technically, we don't check the sign of the integer here,
        # e.g. is_instance_of_common_type(-1, CommonType.UINT) -> True
        return is_integer_type(type)
    else:
        return value_type == type


def promote_numeric_primitive(value: NumericPrimitive, target_type: CommonType) -> NumericPrimitive:
    if target_type == CommonType.INT:
        assert isinstance(value, int), f"cannot promote value {value} to int"
        return value
    elif target_type == CommonType.DOUBLE:
        assert isinstance(value, (int, float)), f"cannot promote value {value} to double"
        return float(value)
    else:
        assert target_type == CommonType.RATIONAL, f"unexpected target type: {target_type}"
        if isinstance(value, Fraction):
            return value
        elif isinstance(value, int):
            return Fraction(value, 1)
        elif isinstance(value, float):
            return Fraction.from_float(value)
        else:
            raise ValueError(f"cannot promote value {value} to rational")


def promote_numeric(value: Numeric, target_type: CommonType) -> Numeric:
    """Promote a numeric value to the target type."""
    if not is_interval_type(target_type):
        assert isinstance(value, NumericPrimitive), f"cannot promote interval {value} to primitive type {target_type}"
        return promote_numeric_primitive(value, target_type)
    if isinstance(value, Interval):
        left = value.left
        right = value.right
    else:
        left = right = value
    base_type = interval_base_type(target_type)
    left = promote_numeric_primitive(left, base_type)
    right = promote_numeric_primitive(right, base_type)
    return Interval(left, right)



"""
(De)serialization of common types.
"""

from typing import no_type_check

import umbi.datatypes
from umbi.datatypes import CommonType, JsonLike, Numeric

from .floats import bytes_to_double, double_to_bytes
from .integers import (
    bytes_to_fixed_size_integer,
    fixed_size_integer_to_bytes,
    num_bytes_for_fixed_size_integer,
)
from .intervals import bytes_to_interval, interval_to_bytes
from .jsons import bytes_to_json, json_to_bytes
from .rationals import bytes_to_rational, rational_to_bytes
from .strings import bytes_to_string, string_to_bytes


def num_bytes_for_common_type(type: CommonType) -> int:
    """Return the number of bytes needed to represent a value of the given type."""
    if umbi.datatypes.is_fixed_size_integer_type(type):
        return num_bytes_for_fixed_size_integer(type)
    if type == CommonType.DOUBLE:
        return 8  # size of double
    if type == CommonType.RATIONAL:
        return num_bytes_for_fixed_size_integer(CommonType.INT64) + num_bytes_for_fixed_size_integer(CommonType.UINT64)
    if umbi.datatypes.is_interval_type(type):
        base_value_type = umbi.datatypes.interval_base_type(type)
        return 2 * num_bytes_for_common_type(base_value_type)
    raise ValueError(f"unsupported common value type: {type}")


@no_type_check
def common_value_to_bytes(value: Numeric | str | JsonLike, type: CommonType, little_endian: bool = True) -> bytes:
    """Convert a value of a given type to a bytestring."""
    assert umbi.datatypes.is_instance_of_common_type(value, type), f"value {value} does not match type {type}"
    if umbi.datatypes.is_fixed_size_integer_type(type):
        return fixed_size_integer_to_bytes(value, type, little_endian)
    elif type == CommonType.DOUBLE:
        return double_to_bytes(value, little_endian=little_endian)
    elif type == CommonType.RATIONAL:
        return rational_to_bytes(value, little_endian=little_endian)
    elif umbi.datatypes.is_interval_type(type):
        return interval_to_bytes(value, type, little_endian=little_endian)
    elif type == CommonType.STRING:
        return string_to_bytes(value)
    elif type == CommonType.JSON:
        return json_to_bytes(value)
    else:
        raise ValueError(f"unsupported common value type: {type}")


def bytes_to_common_value(data: bytes, type: CommonType, little_endian: bool = True) -> Numeric | str | JsonLike:
    """Convert a binary string to a single value of the given common type."""
    if umbi.datatypes.is_fixed_size_integer_type(type):
        return bytes_to_fixed_size_integer(data, type, little_endian=little_endian)
    elif type == CommonType.DOUBLE:
        return bytes_to_double(data, little_endian=little_endian)
    elif type == CommonType.RATIONAL:
        return bytes_to_rational(data, little_endian=little_endian)
    elif umbi.datatypes.is_interval_type(type):
        return bytes_to_interval(data, type, little_endian=little_endian)
    elif type == CommonType.STRING:
        return bytes_to_string(data)
    elif type == CommonType.JSON:
        return bytes_to_json(data)
    else:
        raise ValueError(f"unsupported common value type: {type}")



"""
JSON operations and utilities.
"""

import json as std_json

JsonPrimitive = None | bool | int | float | str
JsonList = list["JsonLike"]
JsonDict = dict[str, "JsonLike"]
JsonLike = JsonPrimitive | JsonList | JsonDict


def is_json_instance(value: object) -> bool:
    if isinstance(value, (type(None), bool, int, float, str)):
        return True
    elif isinstance(value, list):
        return all(is_json_instance(v) for v in value)
    elif isinstance(value, dict):
        return all(isinstance(k, str) and is_json_instance(v) for k, v in value.items())
    return False


def json_remove_none_dict_values(json_obj: JsonLike) -> JsonLike:
    """Recursively remove all None (null) dictionary values from a JSON (sub-)object."""
    if isinstance(json_obj, dict):
        return {k: json_remove_none_dict_values(v) for k, v in json_obj.items() if v is not None}
    elif isinstance(json_obj, list):
        return [json_remove_none_dict_values(v) for v in json_obj]
    return json_obj


def json_to_string(json_obj: JsonLike, indent: int | None = 4, **kwargs) -> str:
    """
    :raises: JSONEncodeError if the object is not serializable
    """
    return std_json.dumps(json_obj, indent=indent, **kwargs)


def string_to_json(json_str: str) -> JsonLike:
    """
    :raises: JSONDecodeError if the string is not valid JSON
    """
    return std_json.loads(json_str)



"""
Interval datatype.
"""

from fractions import Fraction

NumericPrimitive = int | float | Fraction


class Interval:
    """Represents a numeric interval where left <= right."""

    def __init__(self, left: NumericPrimitive, right: NumericPrimitive) -> None:
        self.left = left
        self.right = right
        self.validate()

    def validate(self) -> None:
        if not isinstance(self.left, NumericPrimitive):
            raise ValueError(f"expected numeric left bound, got: {self.left}")
        if not isinstance(self.right, NumericPrimitive):
            raise ValueError(f"expected numeric right bound, got: {self.right}")
        if not self.left <= self.right:
            raise ValueError(f"expected {self.left} <=  {self.right}")

    def __str__(self) -> str:
        return f"interval[{self.left},{self.right}]"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Interval):
            raise ValueError("can only compare Interval with another Interval")
        return self.left == other.left and self.right == other.right

    def __contains__(self, value: NumericPrimitive) -> bool:
        """Check if a numeric value is within the interval."""
        return self.left <= value <= self.right



"""
umbi.binary: Utilities for (de)serializing basic types.
"""

from .common import bytes_to_common_value, common_value_to_bytes
from .sequences import bytes_to_vector, vector_to_bytes

__all__ = [
    "common_value_to_bytes",
    "bytes_to_common_value",
    "bytes_to_vector",
    "vector_to_bytes",
]



"""
umbi.index: handling the .umb index json file.
"""

from .annotations import Annotation, Annotations
from .file_data import FileData
from .model_data import ModelData
from .transition_system import TransitionSystem
from .umb_index import UmbIndex

__all__ = [
    "ModelData",
    "FileData",
    "Annotation",
    "Annotations",
    "TransitionSystem",
    "UmbIndex",
]



"""
Auxiliary vector operations.
"""

from dataclasses import dataclass

from .common_type import CommonType
from .utils import (
    get_instance_type,
    is_instance_of_common_type,
    promote_numeric,
    promote_numeric_primitive,
)

# TODO add CSR vector?
# TODO add vector of ranges (for CSR?)


@dataclass
class VectorType:
    base_type: CommonType


"""Alias for a CSR vector type."""
CSR_TYPE = VectorType(CommonType.UINT64)


def assert_is_list(vector: object):
    if not isinstance(vector, list):
        raise TypeError(f"expected a list/vector, got {type(vector)}")


def is_vector_of_common_type(vector: list, element_type: CommonType) -> bool:
    return all(is_instance_of_common_type(elem, element_type) for elem in vector)


def is_vector_of_type(vector: list, element_type: VectorType) -> bool:
    return is_vector_of_common_type(vector, element_type.base_type)


def promote_to_vector_of_numeric_primitive(vector: list, target_type: CommonType) -> list:
    return [promote_numeric_primitive(elem, target_type) for elem in vector]


def promote_to_vector_of_numeric(vector: list, target_type: CommonType) -> list:
    return [promote_numeric(elem, target_type) for elem in vector]


def vector_element_types(vector: list) -> set[CommonType]:
    """Determine the set of common types of elements in the vector."""
    return set([get_instance_type(x) for x in vector])


def vector_element_type(vector: list) -> CommonType:
    """Determine the common type of elements in the vector. Raises an error if multiple types are found."""
    types = vector_element_types(vector)
    if len(types) != 1:
        raise ValueError(f"vector has multiple element types: {types}")
    return types.pop()


def vector_promotion_type(vector: list) -> CommonType:
    """Determine the common type to which all elements in the vector can be promoted."""
    if len(vector) == 0:
        return CommonType.INT  # whatever
    types = vector_element_types(vector)
    if CommonType.RATIONAL_INTERVAL in types:
        return CommonType.RATIONAL_INTERVAL
    elif CommonType.DOUBLE_INTERVAL in types:
        if CommonType.RATIONAL in types:
            return CommonType.RATIONAL_INTERVAL
        else:
            return CommonType.DOUBLE_INTERVAL
    elif CommonType.RATIONAL in types:
        return CommonType.RATIONAL
    elif CommonType.DOUBLE in types:
        return CommonType.DOUBLE
    elif CommonType.INT in types:
        return CommonType.INT
    else:
        raise ValueError(f"vector contains non-numeric types ({types}), cannot determine promotion type")


def promote_vector(vector: list) -> list:
    """Promote a vector of numeric values to a common type."""
    target_type = vector_promotion_type(vector)
    return promote_to_vector_of_numeric(vector, target_type)


def is_vector_ranges(ranges: list[tuple[int, int]]) -> bool:
    """Check if a vector is a list of ranges."""
    if len(ranges) == 0:
        return False

    def is_interval(x: tuple[int, int]) -> bool:
        return isinstance(x, tuple) and len(x) == 2 and all(isinstance(y, int) for y in x) and x[0] <= x[1]

    if not all(is_interval(interval) for interval in ranges):
        return False
    for i in range(len(ranges) - 1):
        if ranges[i][1] != ranges[i + 1][0]:
            return False
    return True


def is_vector_csr(vector: list[int]) -> bool:
    """Check if a vector is a CSR row start index vector."""
    if len(vector) < 2:
        return False
    if not all(isinstance(x, int) for x in vector):
        return False
    if vector[0] != 0:
        return False
    if not all(vector[i] <= vector[i + 1] for i in range(len(vector) - 1)):
        return False
    return True


def csr_to_ranges(csr: list[int]) -> list[tuple[int, int]]:
    """Convert row start indices to ranges."""
    assert is_vector_csr(csr), "input is not a valid CSR vector"
    ranges = [(csr[i], csr[i + 1]) for i in range(len(csr) - 1)]
    return ranges


def ranges_to_csr(ranges: list[tuple[int, int]]) -> list[int]:
    """Convert ranges to CSR row start indices."""
    assert is_vector_ranges(ranges), "input is not a valid ranges vector"
    csr = [interval[0] for interval in ranges]
    csr.append(ranges[-1][-1])
    return csr



"""
Main UMB index class and schema.
"""

from dataclasses import dataclass, field
from typing import Type

from marshmallow import fields, post_load

import umbi.datatypes

from .annotations import Annotations, AnnotationsSchema
from .file_data import FileData, FileDataSchema
from .json_schema import *
from .model_data import ModelData, ModelDataSchema
from .transition_system import TransitionSystem, TransitionSystemSchema
from .variable_valuations import VariableValuationsSchema


class UmbIndexSchema(JsonSchema):
    """UMB index file schema."""

    format_version = FieldUint(data_key="format-version", required=True)
    format_revision = FieldUint(data_key="format-revision", required=True)
    model_data = fields.Nested(ModelDataSchema, data_key="model-data", required=False)
    file_data = fields.Nested(FileDataSchema, data_key="file-data", required=False)
    transition_system = fields.Nested(TransitionSystemSchema, data_key="transition-system", required=True)
    annotations = fields.Nested(AnnotationsSchema, data_key="annotations", required=False)
    state_valuations = fields.Nested(VariableValuationsSchema, data_key="state-valuations", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> "UmbIndex":
        """Create an UmbIndex object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return UmbIndex(
            format_version=obj.format_version,
            format_revision=obj.format_revision,
            model_data=getattr(obj, "model_data", None),
            file_data=getattr(obj, "file_data", None),
            transition_system=obj.transition_system,
            annotations=getattr(obj, "annotations", None),
            state_valuations=getattr(obj, "state_valuations", None),
        )


@dataclass
class UmbIndex(JsonSchemaResult):

    format_version: int = 0
    format_revision: int = 0
    model_data: ModelData | None = None
    file_data: FileData | None = None
    transition_system: TransitionSystem = field(default_factory=TransitionSystem)
    annotations: Annotations | None = None
    state_valuations: umbi.datatypes.StructType | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return UmbIndexSchema

    def __str__(self) -> str:
        """Convert to a string (json format)."""
        return umbi.datatypes.json_to_string(self.to_json())



"""
Transition system schemas and classes.
"""

from dataclasses import dataclass
from typing import Literal, Type

from marshmallow import fields, post_load, validate

from .json_schema import *


class TransitionSystemSchema(JsonSchema):
    """Transition system schema."""

    time = fields.String(
        data_key="time", required=True, validate=validate.OneOf(["discrete", "stochastic", "urgent-stochastic"])
    )
    num_players = fields.Int(data_key="#players", required=True, validate=validate.Range(min=0))
    num_states = fields.Int(data_key="#states", required=True, validate=validate.Range(min=0))
    num_initial_states = fields.Int(data_key="#initial-states", required=True, validate=validate.Range(min=0))
    num_choices = fields.Int(data_key="#choices", required=True, validate=validate.Range(min=0))
    num_actions = fields.Int(data_key="#actions", required=True, validate=validate.Range(min=0))
    num_branches = fields.Int(data_key="#branches", required=True, validate=validate.Range(min=0))
    num_observations = fields.Int(data_key="#observations", required=False, validate=validate.Range(min=0))

    branch_probability_type = fields.String(
        data_key="branch-probability-type",
        required=False,
        validate=validate.OneOf(["double", "rational", "double-interval", "rational-interval"]),
    )
    exit_rate_type = fields.String(
        data_key="exit-rate-type",
        required=False,
        validate=validate.OneOf(["double", "rational", "double-interval", "rational-interval"]),
    )

    @post_load
    def make_object(self, data: dict, **kwargs) -> "TransitionSystem":
        """Create a TransitionSystem object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return TransitionSystem(
            time=obj.time,
            num_players=obj.num_players,
            num_states=obj.num_states,
            num_initial_states=obj.num_initial_states,
            num_choices=obj.num_choices,
            num_actions=obj.num_actions,
            num_branches=obj.num_branches,
            num_observations=obj.num_observations,
            branch_probability_type=obj.branch_probability_type,
            exit_rate_type=obj.exit_rate_type,
        )


@dataclass
class TransitionSystem(JsonSchemaResult):
    """Transition system data class."""

    time: Literal["discrete", "stochastic", "urgent-stochastic"] = "discrete"
    num_players: int = 0
    num_states: int = 0
    num_initial_states: int = 0
    num_choices: int = 0
    num_actions: int = 0
    num_branches: int = 0
    num_observations: int = 0
    branch_probability_type: Literal["double", "rational", "double-interval", "rational-interval"] | None = None
    exit_rate_type: Literal["double", "rational", "double-interval", "rational-interval"] | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return TransitionSystemSchema



"""
Utilities for packing and unpacking composite datatypes (structs).
"""

from fractions import Fraction

from bitstring import BitArray

from umbi.datatypes import (
    CommonType,
    Numeric,
    StructAttribute,
    StructPadding,
    StructType,
)

from .bitvectors import boolean_pack, boolean_unpack
from .floats import double_pack, double_unpack
from .integers import integer_pack, integer_unpack
from .rationals import rational_pack, rational_unpack
from .strings import string_pack, string_unpack
from .utils import split_bytes


class StructPacker:
    """Utility class for packing structs into a bytestring."""

    def __init__(self):
        self.buffer = BitArray()  # bit buffer, MSB at [0]
        self.bytestring = bytes()  # output bytestring, little-endian order

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def flush_buffer(self):
        """Flush full bytes from the buffer to the bytestring."""
        while len(self.buffer) >= 8:  # append one byte at a time so that the bytestring is little-endian
            self.buffer, bits = self.buffer[:-8], self.buffer[-8:]  # get new bits from the end (LSB side)
            new_bytes = bits.tobytes()
            self.bytestring += new_bytes

    def append_to_buffer(self, bits: BitArray):
        """Append bits to the buffer flush full bytes to the bytestring."""
        self.buffer = bits + self.buffer  # prepend to the start (MSB side)
        self.flush_buffer()

    def add_padding(self, num_bits: int):
        """Add padding bits to the buffer."""
        assert num_bits > 0
        self.append_to_buffer(BitArray(uint=0, length=num_bits))

    def pack_attribute(self, field: StructAttribute, value: object):
        """Pack a single attribute into the buffer or the bytestring."""
        if field.type in [CommonType.STRING, CommonType.RATIONAL]:
            # expected to be byte-aligned before packing string/rational
            self.assert_buffer_empty()
            # send byte-aligned output directly to the bytestring
            if field.type == CommonType.STRING:
                assert isinstance(value, str)
                self.bytestring += string_pack(value)
            else:  # field.type == CommonType.RATIONAL:
                assert isinstance(value, Fraction)
                self.bytestring += rational_pack(value)
            return
        # the remaining datatypes are bitstrings that go into the buffer
        value_bits = None
        if field.type == CommonType.BOOLEAN:
            assert isinstance(value, bool)
            value_bits = boolean_pack(value, field.size)
        elif field.type in [CommonType.INT, CommonType.UINT]:
            assert isinstance(value, int)
            signed = field.type == CommonType.INT
            assert field.size is not None
            value_bits = integer_pack(value, field.size, signed)
        elif field.type == CommonType.DOUBLE:
            assert isinstance(value, float)
            assert field.size == 64, f"expected {field.size} to be 64 for double type"
            value_bits = double_pack(value)
        else:
            raise ValueError(f"unsupported field type: {field.type}")
        self.append_to_buffer(value_bits)

    def pack_struct(self, value_type: StructType, values: dict[str, object]) -> bytes:
        for field in value_type:
            if isinstance(field, StructPadding):
                self.add_padding(field.padding)
                continue
            assert isinstance(field, StructAttribute)
            self.pack_attribute(field, values[field.name])
        self.assert_buffer_empty()
        return self.bytestring


class StructUnpacker:
    """Utility class for unpacking composite datatypes from a bytestring."""

    def __init__(self, bytestring: bytes):
        self.bytestring = bytestring  # input bytestring, little-endian order
        self.buffer = BitArray()  # bit buffer, MSB at [0]

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def align_buffer(self, num_bits: int):
        """Ensure the buffer has the required number ofbits."""
        if len(self.buffer) >= num_bits:
            return
        num_bytes_needed = (num_bits - len(self.buffer) + 7) // 8
        assert len(self.bytestring) >= num_bytes_needed, "not enough data to fill the buffer"
        for _ in range(num_bytes_needed):  # read one byte at a time since the bytestring is little-endian
            next_bytes, self.bytestring = split_bytes(self.bytestring, 1)
            self.buffer = BitArray(bytes=next_bytes) + self.buffer  # prepend to the buffer to keep MSB at [0]

    def extract_from_buffer(self, num_bits: int):
        """Extract the given number of bits from the buffer."""
        assert num_bits >= 0
        self.align_buffer(num_bits)
        bits = self.buffer[-num_bits:]  # remove from the end (LSB side)
        self.buffer = self.buffer[:-num_bits]
        return bits

    def skip_padding(self, num_bits: int):
        self.extract_from_buffer(num_bits)

    def unpack_attribute(self, field: StructAttribute) -> str | bool | Numeric:
        """Unpack a single field from the buffer or the bytestring."""
        if field.type in [CommonType.STRING, CommonType.RATIONAL]:
            self.assert_buffer_empty()
            if field.type == CommonType.STRING:
                value, self.bytestring = string_unpack(self.bytestring)
                return value
            elif field.type == CommonType.RATIONAL:
                value, self.bytestring = rational_unpack(self.bytestring)
                return value
        # fixed-size types come from the buffer
        assert field.size is not None
        bits = self.extract_from_buffer(field.size)
        if field.type == CommonType.BOOLEAN:
            return boolean_unpack(bits)
        elif field.type in [CommonType.INT, CommonType.UINT]:
            signed = field.type == CommonType.INT
            return integer_unpack(bits, signed)
        elif field.type == CommonType.DOUBLE:
            return double_unpack(bits)
        else:
            raise ValueError(f"unsupported field type: {field.type}")

    def unpack_struct(self, value_type: StructType) -> dict[str, object]:
        name_value = dict()
        for field in value_type:
            if isinstance(field, StructPadding):
                self.skip_padding(field.padding)
                continue
            name_value[field.name] = self.unpack_attribute(field)
        self.assert_buffer_empty()
        return name_value


def struct_pack(value_type: StructType, values: dict[str, object]) -> bytes:
    """Convert a composite datatype to a BitArray."""
    return StructPacker().pack_struct(value_type, values)


def struct_unpack(bytestring: bytes, value_type: StructType) -> dict[str, object]:
    """Unpack a BitArray to a composite datatype."""
    return StructUnpacker(bytestring).unpack_struct(value_type)



"""
Data types and type identifiers used throughout umbi. This module centralizes:
1. String identifiers for common value types in umbfiles
2. Python type aliases for these common types
3. Functions to check and assert type names and instances
"""

import enum


class CommonType(str, enum.Enum):
    """String literals for common datatypes."""

    BYTES = "bytes"
    BOOLEAN = "bool"

    INT = "int"
    UINT = "uint"

    INT16 = "int16"
    UINT16 = "uint16"
    INT32 = "int32"
    UINT32 = "uint32"
    INT64 = "int64"
    UINT64 = "uint64"

    DOUBLE = "double"
    RATIONAL = "rational"

    DOUBLE_INTERVAL = "double-interval"
    RATIONAL_INTERVAL = "rational-interval"

    STRING = "string"
    JSON = "json"

    STRUCT = "struct"


### integers


def is_fixed_size_integer_type(type: CommonType) -> bool:
    return type in [
        CommonType.INT16,
        CommonType.UINT16,
        CommonType.INT32,
        CommonType.UINT32,
        CommonType.INT64,
        CommonType.UINT64,
    ]


def is_variable_size_integer_type(type: CommonType) -> bool:
    return type in [CommonType.INT, CommonType.UINT]


def is_integer_type(type: CommonType) -> bool:
    return is_variable_size_integer_type(type) or is_fixed_size_integer_type(type)


def assert_integer_type(type: CommonType):
    assert is_integer_type(type), f"not an integer type: {type}"


def integer_type_signed(type: CommonType) -> bool:
    assert_integer_type(type)
    return type.name.startswith("i")


### interval types (defined in interval.py)


def is_interval_type(type: CommonType) -> bool:
    return type in [CommonType.DOUBLE_INTERVAL, CommonType.RATIONAL_INTERVAL]


def assert_interval_type(type: CommonType):
    assert is_interval_type(type), f"not an interval type: {type}"


def interval_base_type(type: CommonType) -> CommonType:
    assert_interval_type(type)
    return {
        CommonType.DOUBLE_INTERVAL: CommonType.DOUBLE,
        CommonType.RATIONAL_INTERVAL: CommonType.RATIONAL,
    }[type]


# def is_numeric_type(type: CommonType) -> bool:
#     """Check if the given common type is a numeric type (including intervals)."""
#     return is_integer_type(type) or type in [
#         CommonType.DOUBLE,
#         CommonType.RATIONAL,
#         CommonType.DOUBLE_INTERVAL,
#         CommonType.RATIONAL_INTERVAL
#     ]



from dataclasses import dataclass, field
from enum import Enum
from typing import Iterable

from umbi.datatypes import (
    CommonType,
    Interval,
    Numeric,
    StructType,
    vector_promotion_type,
)

from .annotation import AtomicPropositionAnnotation, RewardAnnotation
from .model_info import ModelInfo


class TimeType(str, Enum):
    """Time types for transition systems."""

    DISCRETE = "discrete"
    STOCHASTIC = "stochastic"
    URGENT_STOCHASTIC = "urgent-stochastic"


@dataclass
class ExplicitAts:
    """Explicit container for an annotated transition system (ATS)."""

    # TODO annotate fields, describe optionality and defaults

    model_info: ModelInfo | None = None
    time: TimeType = TimeType.DISCRETE
    num_states: int = 0

    num_players: int = 0
    state_to_player: list[int] | None = None

    num_initial_states: int = 0
    state_is_initial: list[bool] = field(default_factory=list)

    num_observations: int = 0

    num_choices: int = 0
    state_to_choice: list[int] | None = None

    state_is_markovian: list[bool] | None = None
    state_exit_rate: list[Numeric] | None = None

    num_actions: int = 1
    choice_to_action: list[int] | None = None
    action_strings: list[str] | None = None

    num_branches: int = 0
    choice_to_branch: list[int] | None = None
    branch_to_target: list[int] | None = None
    branch_probabilities: list[Numeric] | None = None

    reward_annotations: dict[str, RewardAnnotation] = field(default_factory=dict)
    ap_annotations: dict[str, AtomicPropositionAnnotation] = field(default_factory=dict)

    # TODO consolidate into one structure
    state_valuations: StructType | None = None
    state_valuations_values: list[dict] | None = None

    def __eq__(self, other):
        # TODO implement
        if not isinstance(other, ExplicitAts):
            return False
        if self.num_states != other.num_states:
            return False
        if self.num_players != other.num_players:
            return False
        if self.num_actions != other.num_actions:
            return False
        return True

    @property
    def initial_states(self) -> list[int]:
        """Get the list of the initial states."""
        return [i for i, is_initial in enumerate(self.state_is_initial) if is_initial]

    def set_initial_states(self, initial_states: list[int]):
        """Set the initial states."""
        self.state_is_initial = [False] * self.num_states
        for s in initial_states:
            if s >= self.num_states:
                raise ValueError(f"Invalid initial state {s}, must be < {self.num_states}.")
            self.state_is_initial[s] = True

    @property
    def markovian_states(self) -> list[int]:
        """Get the list of the markovian states."""
        if self.state_is_markovian is None:
            raise ValueError("state_is_markovian is not set")
        return [i for i, is_markovian in enumerate(self.state_is_markovian) if is_markovian]

    def get_branch_probability(self, branch_id: int) -> Numeric:
        if self.branch_probabilities is not None:
            return self.branch_probabilities[branch_id]
        return 1.0

    def get_branch_target(self, branch_id: int) -> int:
        if self.branch_to_target is not None:
            return self.branch_to_target[branch_id]
        raise RuntimeError("Branches must have targets in UMBI.")

    def get_choice_action(self, choice_id: int) -> int:
        """Get the action identifier for a choice"""
        if self.choice_to_action is not None:
            return self.choice_to_action[choice_id]
        raise NotImplementedError("Getting actions when not set are not implemented.")
        # TODO what is the default here?

    def get_action_name(self, action_id: int) -> str:
        """Get the name of an action identifier"""
        if self.action_strings is not None:
            return self.action_strings[action_id]
        raise ValueError("action strings are not set")

    def state_choice_range(self, state: int) -> Iterable[int]:
        """Return the choice range of the given state."""
        if self.state_to_choice is not None:
            assert state <= self.num_states
            return range(self.state_to_choice[state], self.state_to_choice[state + 1])
        else:
            return range(state, state + 1)

    def choice_branch_range(self, choice: int) -> Iterable[int]:
        """Return the branch range of the given choice."""
        if self.choice_to_branch is not None:
            assert choice <= self.num_choices
            return range(self.choice_to_branch[choice], self.choice_to_branch[choice + 1])
        else:
            return range(choice, choice + 1)

    @property
    def branch_probability_type(self) -> CommonType | None:
        if self.branch_probabilities is None:
            return None
        return vector_promotion_type(self.branch_probabilities)

    @property
    def exit_rate_type(self) -> CommonType | None:
        if self.state_exit_rate is None:
            return None
        return vector_promotion_type(self.state_exit_rate)

    @property
    def reward_annotation_names(self) -> list[str]:
        """Get the names of all reward annotations."""
        return list(self.reward_annotations.keys())

    def add_reward_annotation(self, annotation: RewardAnnotation):
        """Add a reward annotation."""
        if annotation.name in self.reward_annotations:
            raise ValueError(f"reward annotation with name {annotation.name} already exists")
        self.reward_annotations[annotation.name] = annotation

    def get_reward_annotation(self, name: str) -> RewardAnnotation:
        """Get the reward annotation with the given name."""
        if name not in self.reward_annotations:
            raise ValueError(f"reward annotation with name {name} does not exist")
        return self.reward_annotations[name]

    @property
    def ap_annotation_names(self) -> list[str]:
        """Get the names of all atomic proposition annotations."""
        return list(self.ap_annotations.keys())

    def add_ap_annotation(self, annotation: AtomicPropositionAnnotation):
        """Add an atomic proposition annotation."""
        if annotation.name in self.ap_annotations:
            raise ValueError(f"AP annotation with name {annotation.name} already exists")
        self.ap_annotations[annotation.name] = annotation

    def get_ap_annotation(self, name: str) -> AtomicPropositionAnnotation:
        """Get the atomic proposition annotation with the given name."""
        if name not in self.ap_annotations:
            raise ValueError(f"AP annotation with name {name} does not exist")
        return self.ap_annotations[name]

    @property
    def has_state_valuations(self) -> bool:
        return self.state_valuations is not None

    def validate(self):
        # TODO implement

        if not self.num_states > 0:
            raise ValueError("expected num_states > 0")

        if self.num_players > 1:
            if self.state_to_player is None:
                raise ValueError("num_players > 1 but state_to_player is None")
            if len(self.state_to_player) != self.num_states:
                raise ValueError("expected len(state_to_player) == num_states")

        if self.num_initial_states != len(self.initial_states):
            raise ValueError("expected num_initial_states == len(initial_states)")

        if self.state_to_choice is not None:
            if len(self.state_to_choice) != self.num_states + 1:
                raise ValueError("expected len(state_to_choice) == num_states+1")
        if self.choice_to_branch is not None:
            if len(self.choice_to_branch) != self.num_choices + 1:
                raise ValueError("expected len(choice_to_branch) == num_choices+1")
        if self.branch_to_target is not None:
            if len(self.branch_to_target) != self.num_branches:
                raise ValueError("expected len(branch_to_target) == num_branches")
        if self.has_state_valuations:
            if self.state_valuations_values is None:
                raise ValueError("state_valuations is set but state_valuations_values is None")
            if len(self.state_valuations_values) != self.num_states:
                raise ValueError("expected len(state_valuations_values) == num_states")



"""
umbi.datatypes: Core datatype definitions and utilities.

This package contains the fundamental datatype classes and definitions used throughout umbi,
including intervals, composite types, vector utilities, and JSON operations. Serialization operations remain in umbi.binary.
"""

from .common_type import *
from .interval import Interval
from .json import *
from .struct import *
from .utils import *
from .vector import *

__all__ = [
    # definitions.py
    "CommonType",
    "is_fixed_size_integer_type",
    "is_variable_size_integer_type",
    "is_integer_type",
    "is_interval_type",
    "assert_interval_type",
    "interval_base_type",
    # interval.py
    "Interval",
    # struct.py
    "StructPadding",
    "StructAttribute",
    "StructType",
    # json.py
    "JsonPrimitive",
    "JsonList",
    "JsonDict",
    "JsonLike",
    "is_json_instance",
    "json_remove_none_dict_values",
    "json_to_string",
    "string_to_json",
    # utils.py
    "NumericPrimitive",
    "Numeric",
    "is_instance_of_common_type",
    "promote_numeric_primitive",
    "promote_numeric",
    # vector.py
    "VectorType",
    "CSR_TYPE",
    "assert_is_list",
    "is_vector_of_common_type",
    "is_vector_of_type",
    "vector_promotion_type",
    "promote_vector",
]



"""
umbi.io: Utilities for binary and JSON serialization, and tar archive handling.
"""

from . import index
from .umb import ExplicitUmb, read_umb, write_umb
from .umb_ats_converter import read_ats, write_ats

__all__ = [
    "index",
    "ExplicitUmb",
    "read_umb",
    "write_umb",
    "read_ats",
    "write_ats",
]



from enum import Enum

from umbi.datatypes import CommonType, Numeric


class Annotation:
    """General annotation."""

    # TODO add lower: float | None = None
    # TODO add upper: float | None = None

    def __init__(
        self,
        name: str,
        type: CommonType,
        alias: str | None = None,
        description: str | None = None,
        state_to_value: list[Numeric] | None = None,
        choice_to_value: list[Numeric] | None = None,
        branch_to_value: list[Numeric] | None = None,
    ):
        self._name = name
        if not type in [
            CommonType.BOOLEAN,
            CommonType.STRING,
            CommonType.DOUBLE,
            CommonType.RATIONAL,
            CommonType.DOUBLE_INTERVAL,
            CommonType.RATIONAL_INTERVAL,
        ]:
            raise ValueError(f"Invalid annotation type: {type}")
        self._type = type
        self.alias = alias
        self.description = description
        self.state_to_value = state_to_value
        self.choice_to_value = choice_to_value
        self.branch_to_value = branch_to_value

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> CommonType:
        return self._type

    def validate(self) -> None:
        """Validate the annotation data."""
        # TODO check datatypes
        # TODO check lower
        # TODO check upper
        pass


class RewardAnnotation(Annotation):
    """Reward annotation data class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.type not in [
            CommonType.DOUBLE,
            CommonType.RATIONAL,
            CommonType.DOUBLE_INTERVAL,
            CommonType.RATIONAL_INTERVAL,
        ]:
            raise ValueError(f"Invalid reward annotation type: {self.type}")


class AtomicPropositionAnnotation(Annotation):
    """Atomic proposition annotation data class."""

    def __init__(
        self,
        name: str,
        alias: str | None = None,
        description: str | None = None,
        state_to_value: list[Numeric] | None = None,
        choice_to_value: list[Numeric] | None = None,
        branch_to_value: list[Numeric] | None = None,
    ):
        super().__init__(
            name=name,
            type=CommonType.BOOLEAN,
            alias=alias,
            description=description,
            state_to_value=state_to_value,
            choice_to_value=choice_to_value,
            branch_to_value=branch_to_value,
        )
        if self.type != CommonType.BOOLEAN:
            raise ValueError(f"Invalid atomic proposition annotation type: {self.type}")



"""
(De)serialization of jsons.
"""

import umbi.datatypes

from .strings import bytes_to_string, string_to_bytes


def bytes_to_json(data: bytes) -> umbi.datatypes.JsonLike:
    """Convert bytes to a JSON object."""
    return umbi.datatypes.string_to_json(bytes_to_string(data))


def json_to_bytes(json_obj: umbi.datatypes.JsonLike) -> bytes:
    """Convert a JSON object to bytes."""
    return string_to_bytes(umbi.datatypes.json_to_string(json_obj))



"""
(De)serialization of sequences of common types or structs.
"""

import logging

from umbi.datatypes import CommonType, StructType

from .bitvectors import bitvector_to_bytes, bytes_to_bitvector
from .common import (
    bytes_to_common_value,
    common_value_to_bytes,
    num_bytes_for_common_type,
)
from .structs import struct_pack, struct_unpack

logger = logging.getLogger(__name__)


def bytes_into_chunk_ranges(data: bytes, chunk_ranges: list[tuple[int, int]]) -> list[bytes]:
    """Split bytestring into chunks according to chunk ranges."""
    assert len(data) == chunk_ranges[-1][1], "data length does not match the end of the last chunk range"
    return [data[start:end] for start, end in chunk_ranges]


def chunks_to_csr(chunks: list[bytes]) -> list[int]:
    """Build csr for a list of chunks."""
    current_pos = 0
    csr = [current_pos]
    for chunk in chunks:
        current_pos += len(chunk)
        csr.append(current_pos)
    return csr


def bytes_into_chunks(data: bytes, chunk_size: int) -> list[bytes]:
    """Split bytestring into evenly sized chunks."""
    assert chunk_size > 0, f"expected {chunk_size} to be a positive number"
    assert len(data) % chunk_size == 0, f"expected {len(data)} to be divisible by {chunk_size}"
    return [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]


def bytes_to_vector(
    data: bytes,
    value_type: CommonType | StructType,
    chunk_ranges: list[tuple[int, int]] | None = None,
    little_endian: bool = True,
) -> list:
    """
    Decode a binary string as a list of numbers.
    :param value_type: vector element type, either composite, bool, string or one of {int32|uint32|int64|uint64|double|rational}[-interval]
    :param chunk_ranges: (optional) chunk ranges to split the data into
    :param little_endian: if True, the binary string is interpreted as little-endian
    """

    if isinstance(value_type, StructType):
        assert chunk_ranges is not None, "chunk_ranges must be provided when value_type is a StructType"
        chunks = bytes_into_chunk_ranges(data, chunk_ranges)
        return [struct_unpack(chunk, value_type) for chunk in chunks]

    if value_type == CommonType.BOOLEAN:
        assert little_endian, "big-endianness for bitvectors is not implemented"
        return bytes_to_bitvector(data)

    if len(data) == 0:
        return []

    if chunk_ranges is None:
        chunk_size = num_bytes_for_common_type(value_type)
        chunks = bytes_into_chunks(data, chunk_size)
    else:
        chunks = bytes_into_chunk_ranges(data, chunk_ranges)
    return [bytes_to_common_value(chunk, value_type, little_endian) for chunk in chunks]


def vector_to_bytes(
    vector: list, value_type: CommonType | StructType, little_endian: bool = True
) -> tuple[bytes, list[int] | None]:
    """Encode a list of values as a binary string.
    :param value_type: vector element type, either composite, bool, string or {int32|uint32|int64|uint64|double|rational}[-interval]
    :return: encoded binary string
    :return: (optional) chunk csr if non-trivial splitting is needed to split the resulting bytestring into chunks, e.g. for strings or non-standard rationals
    """

    if len(vector) == 0:
        logger.warning("converting empty vector to bytes")
        return (b"", None)

    if isinstance(value_type, StructType):
        chunks = [struct_pack(value_type, item) for item in vector]
        chunks_csr = chunks_to_csr(chunks)
        bytestring = b"".join(chunks)
        return bytestring, chunks_csr

    if value_type == CommonType.BOOLEAN:
        assert little_endian, "big-endianness for bitvectors is not implemented"
        return (bitvector_to_bytes(vector), None)

    chunks = [common_value_to_bytes(item, value_type, little_endian) for item in vector]
    chunks_csr = None
    if value_type == CommonType.STRING or any(len(chunk) != num_bytes_for_common_type(value_type) for chunk in chunks):
        chunks_csr = chunks_to_csr(chunks)
    bytestring = b"".join(chunks)

    return bytestring, chunks_csr



"""
Composite datatype definitions.

This module contains the core classes for defining composite datatypes (structs).
The serialization operations for composites remain in umbi.binary.composites.
"""

from dataclasses import dataclass, field

from .common_type import *


@dataclass
class StructPadding:
    """Padding bits in a composite datatype."""

    padding: int

    def validate(self):
        if self.padding <= 0:
            raise ValueError(f"Padding must be positive ({self.padding})")


@dataclass
class StructAttribute:
    """A variable field in a composite datatype."""

    name: str
    type: CommonType
    size: int | None = (
        None  # number of bits for values of fixed-size types; None for variable-size types (string, rational)
    )
    lower: float | None = None  # lower bound (for numeric types)
    upper: float | None = None  # upper bound (for numeric types)
    offset: float | None = None  # lower value offset (for numeric types)

    def validate(self):
        if self.type not in [
            CommonType.BOOLEAN,
            CommonType.INT,
            CommonType.UINT,
            CommonType.DOUBLE,
            CommonType.RATIONAL,
            CommonType.STRING,
        ]:
            raise ValueError(f"Unsupported field type: {self.type}")
        if self.size is None:
            if self.type in [CommonType.INT, CommonType.UINT]:
                raise ValueError("Field size must be specified for fixed-size types (int,uint)")
        else:
            if self.size <= 0:
                raise ValueError(f"Field size must be positive ({self.size})")
            elif self.type == CommonType.DOUBLE and self.size != 64:
                raise ValueError("Field size for double must be 64")


@dataclass
class StructType:
    """A composite datatype consisting of attributes and paddings."""

    alignment: int  # alignment in bits
    fields: list[StructPadding | StructAttribute] = field(default_factory=list)

    # @classmethod
    # def from_namespace(cls, data: SimpleNamespace) -> "StructType":
    #     fields = []
    #     for item in data.variables:
    #         if hasattr(item, "padding"):
    #             fields.append(StructPadding.from_namespace(item))
    #         else:
    #             fields.append(StructAttribute.from_namespace(item))
    #     return cls(data.alignment, fields)

    def validate(self):
        for item in self.fields:
            item.validate()
        # TODO check alignment

    def __str__(self) -> str:
        lines = [f"struct (alignment={self.alignment}):"]
        for f in self.fields:
            lines.append(f"  {f}")
        return "\n".join(lines)

    def __iter__(self):
        return iter(self.fields)



import os
import tomllib

from . import ats, binary, datatypes, io
from .logger import set_log_level, setup_logging
from .version import __format_revision__, __format_version__, __toolname__, __version__

__all__ = [
    "set_log_level",
    "setup_logging",
    "__toolname__",
    "__version__",
    "__format_version__",
    "__format_revision__",
    "datatypes",
    "binary",
    "io",
    "ats",
]



import time

import umbi
import umbi.ats
from umbi.datatypes import (
    CommonType,
    promote_to_vector_of_numeric,
    vector_promotion_type,
)

from .index import (
    Annotation,
    Annotations,
    FileData,
    ModelData,
    TransitionSystem,
    UmbIndex,
)
from .umb import ExplicitUmb, read_umb, write_umb


def umbi_file_data() -> FileData:
    return FileData(
        tool=umbi.version.__toolname__,
        tool_version=umbi.version.__version__,
        creation_date=int(time.time()),
        # parameters=parameters,
    )


def umb_annotation_to_ats_annotation(
    umb_annotation: Annotation, applies_to_values: dict[str, list], ats_annotation: umbi.ats.Annotation
):
    for applies_to in umb_annotation.applies_to or []:
        if applies_to == "states":
            ats_annotation.state_to_value = applies_to_values["states"]
        elif applies_to == "choices":
            ats_annotation.choice_to_value = applies_to_values["choices"]
        elif applies_to == "branches":
            ats_annotation.branch_to_value = applies_to_values["branches"]


def ats_annotation_to_umb_annotation(annotation: umbi.ats.Annotation) -> tuple[Annotation, dict[str, list]]:
    applies_to = []
    applies_to_values = {}
    target_type = annotation.type
    for key, values in [
        ("states", annotation.state_to_value),
        ("choices", annotation.choice_to_value),
        ("branches", annotation.branch_to_value),
    ]:
        if values is not None:
            applies_to.append(key)
            if target_type not in [CommonType.STRING, CommonType.BOOLEAN]:
                # implicit promotion
                values = promote_to_vector_of_numeric(values, target_type)
            applies_to_values[key] = values
    umb_annotation = Annotation(
        alias=annotation.alias,
        description=annotation.description,
        applies_to=applies_to,
        type=annotation.type.value,
        lower=None,
        upper=None,
    )
    return umb_annotation, applies_to_values


def explicit_umb_to_explicit_ats(umb: ExplicitUmb) -> umbi.ats.ExplicitAts:
    umb.validate()
    ats = umbi.ats.ExplicitAts()

    ## index
    # skip format_version
    # skip format_revision
    md = umb.index.model_data
    if md is not None:
        ats.model_info = umbi.ats.ModelInfo(
            name=md.name,
            version=md.version,
            authors=md.authors,
            description=md.description,
            comment=md.comment,
            doi=md.doi,
            url=md.url,
        )
    # skip file_data
    # load index.transition_system fields into ats
    ts = umb.index.transition_system
    ats.time = umbi.ats.TimeType(ts.time)
    ats.num_players = ts.num_players
    ats.num_states = ts.num_states
    ats.num_initial_states = ts.num_initial_states
    ats.num_choices = ts.num_choices
    ats.num_actions = ts.num_actions
    ats.num_branches = ts.num_branches
    ats.num_observations = ts.num_observations

    # load annotations
    if umb.index.annotations is not None:
        if umb.index.annotations.rewards is not None:
            assert umb.rewards is not None
            for name, umb_annotation in umb.index.annotations.rewards.items():
                assert name in umb.rewards
                applies_to_values = umb.rewards[name]
                ats_annotation = umbi.ats.RewardAnnotation(
                    name=name,
                    type=CommonType(umb_annotation.type) if umb_annotation.type is not None else CommonType.BOOLEAN,
                    alias=umb_annotation.alias,
                    description=umb_annotation.description,
                )
                umb_annotation_to_ats_annotation(umb_annotation, applies_to_values, ats_annotation)
                ats.add_reward_annotation(ats_annotation)
        if umb.index.annotations.aps is not None:
            assert umb.aps is not None
            for name, umb_annotation in umb.index.annotations.aps.items():
                assert name in umb.aps
                applies_to_values = umb.aps[name]
                ats_annotation = umbi.ats.AtomicPropositionAnnotation(
                    name=name,
                    alias=umb_annotation.alias,
                    description=umb_annotation.description,
                )
                umb_annotation_to_ats_annotation(umb_annotation, applies_to_values, ats_annotation)
                ats.add_ap_annotation(ats_annotation)

    ## values
    ats.state_is_initial = umb.state_is_initial
    ats.state_to_choice = umb.state_to_choice
    ats.state_to_player = umb.state_to_player

    ats.state_is_markovian = umb.state_is_markovian
    ats.state_exit_rate = umb.state_exit_rate

    ats.choice_to_branch = umb.choice_to_branch
    ats.choice_to_action = umb.choice_to_action
    ats.action_strings = umb.action_strings

    ats.branch_to_target = umb.branch_to_target
    ats.branch_probabilities = umb.branch_probabilities

    # TODO consolidate into one structure
    ats.state_valuations = umb.index.state_valuations
    ats.state_valuations_values = umb.state_valuations

    ats.validate()
    return ats


def explicit_ats_to_explicit_umb(ats: umbi.ats.ExplicitAts) -> ExplicitUmb:
    ats.validate()
    umb = ExplicitUmb()

    ## index
    # insert our format_version, format_revision

    reward_annotations = {}
    reward_values = {}
    for name, ats_annotation in ats.reward_annotations.items():
        umb_annotation, applies_to_values = ats_annotation_to_umb_annotation(ats_annotation)
        reward_values[name] = applies_to_values
        reward_annotations[name] = umb_annotation
    ap_annotations = {}
    ap_values = {}
    for name, ats_annotation in ats.ap_annotations.items():
        umb_annotation, applies_to_values = ats_annotation_to_umb_annotation(ats_annotation)
        ap_values[name] = applies_to_values
        ap_annotations[name] = umb_annotation

    umb.index = UmbIndex(
        format_version=umbi.version.__format_version__,
        format_revision=umbi.version.__format_revision__,
        # create transition_system from matching fields
        transition_system=TransitionSystem(
            time=ats.time.value,
            num_players=ats.num_players,
            num_states=ats.num_states,
            num_initial_states=ats.num_initial_states,
            num_choices=ats.num_choices,
            num_actions=ats.num_actions,
            num_branches=ats.num_branches,
            num_observations=ats.num_observations,
            branch_probability_type=None,  # for now
            exit_rate_type=None,  # for now
        ),
        model_data=(
            ModelData(
                name=ats.model_info.name,
                version=ats.model_info.version,
                authors=ats.model_info.authors,
                description=ats.model_info.description,
                comment=ats.model_info.comment,
                doi=ats.model_info.doi,
                url=ats.model_info.url,
            )
            if ats.model_info is not None
            else None
        ),
        # insert our file_data
        file_data=umbi_file_data(),
        # create annotations
        annotations=Annotations(
            rewards=reward_annotations,
            aps=ap_annotations,
        ),
        state_valuations=ats.state_valuations,
    )
    ## values
    umb.state_is_initial = ats.state_is_initial
    umb.state_to_choice = ats.state_to_choice
    umb.state_to_player = ats.state_to_player

    umb.state_is_markovian = ats.state_is_markovian
    if ats.state_exit_rate is not None:
        # derive promotion type and promote
        target_type = ats.exit_rate_type
        assert target_type is not None
        umb.index.exit_rate_type = target_type.value
        umb.state_exit_rate = promote_to_vector_of_numeric(ats.state_exit_rate, target_type)

    umb.choice_to_branch = ats.choice_to_branch
    umb.choice_to_action = ats.choice_to_action
    umb.action_strings = ats.action_strings

    umb.branch_to_target = ats.branch_to_target
    if ats.branch_probabilities is not None:
        # derive promotion type and promote
        target_type = ats.branch_probability_type
        assert target_type is not None
        umb.index.branch_probability_type = target_type.value
        umb.branch_probabilities = promote_to_vector_of_numeric(ats.branch_probabilities, target_type)

    umb.rewards = reward_values
    umb.aps = ap_values
    umb.state_valuations = ats.state_valuations_values

    umb.validate()
    return umb


def read_ats(umbpath: str) -> umbi.ats.ExplicitAts:
    """Read ATS from a umbfile."""
    umb = read_umb(umbpath)
    ats = explicit_umb_to_explicit_ats(umb)
    return ats


def write_ats(ats: umbi.ats.ExplicitAts, umbpath: str):
    """Write ATS to a umbfile."""
    umb = explicit_ats_to_explicit_umb(ats)
    write_umb(umb, umbpath)



"""
File data schemas and classes.
"""

from dataclasses import dataclass
from typing import Type

from marshmallow import fields, post_load

import umbi.datatypes

from .json_schema import *


class FileDataSchema(JsonSchema):
    """File data schema."""

    tool = fields.String(data_key="tool", required=False)
    tool_version = fields.String(data_key="tool-version", required=False)
    creation_date = FieldUint(data_key="creation-date", required=False)
    parameters = fields.Raw(data_key="parameters", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> "FileData":
        """Create a FileData object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return FileData(
            tool=obj.tool,
            tool_version=obj.tool_version,
            creation_date=obj.creation_date,
            parameters=obj.parameters,
        )


@dataclass
class FileData(JsonSchemaResult):
    """File data class."""

    tool: str | None = None
    tool_version: str | None = None
    creation_date: int | None = None
    parameters: umbi.datatypes.JsonLike | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return FileDataSchema
